/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_xerial_snappy_BitShuffleNative */

#ifndef _Included_org_xerial_snappy_BitShuffleNative
#define _Included_org_xerial_snappy_BitShuffleNative
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_xerial_snappy_BitShuffleNative
 * Method:    shuffle
 * Signature: (Ljava/lang/Object;IIILjava/lang/Object;I)I
 */
JNIEXPORT jint JNICALL Java_org_xerial_snappy_BitShuffleNative_shuffle
  (JNIEnv *, jobject, jobject, jint, jint, jint, jobject, jint);

/*
 * Class:     org_xerial_snappy_BitShuffleNative
 * Method:    shuffleDirectBuffer
 * Signature: (Ljava/nio/ByteBuffer;IIILjava/nio/ByteBuffer;I)I
 */
JNIEXPORT jint JNICALL Java_org_xerial_snappy_BitShuffleNative_shuffleDirectBuffer
  (JNIEnv *, jobject, jobject, jint, jint, jint, jobject, jint);

/*
 * Class:     org_xerial_snappy_BitShuffleNative
 * Method:    unshuffle
 * Signature: (Ljava/lang/Object;IIILjava/lang/Object;I)I
 */
JNIEXPORT jint JNICALL Java_org_xerial_snappy_BitShuffleNative_unshuffle
  (JNIEnv *, jobject, jobject, jint, jint, jint, jobject, jint);

/*
 * Class:     org_xerial_snappy_BitShuffleNative
 * Method:    unshuffleDirectBuffer
 * Signature: (Ljava/nio/ByteBuffer;IIILjava/nio/ByteBuffer;I)I
 */
JNIEXPORT jint JNICALL Java_org_xerial_snappy_BitShuffleNative_unshuffleDirectBuffer
  (JNIEnv *, jobject, jobject, jint, jint, jint, jobject, jint);

#ifdef __cplusplus
}
#endif
#endif
